@@ -6,10 +6,7 @@ use std::rc::Rc;
 use std::str::FromStr;
 use std::collections::HashMap;
 
-use log::warn;
-
-use crate::tree;
-use crate::{Rect, Error};
+use crate::{Rect, Error, EnableBackground};
 use super::{Document, Attribute, AId, EId, Node, NodeId, NodeKind, NodeData, AttributeValue};
 
 const SVG_NS: &str = "http://www.w3.org/2000/svg";
@@ -22,7 +19,7 @@ impl Document {
         parse(xml)
     }
 
-    fn append(&mut self, parent_id: NodeId, kind: NodeKind) -> NodeId {
+    pub(super) fn append(&mut self, parent_id: NodeId, kind: NodeKind) -> NodeId {
         let new_child_id = NodeId(self.nodes.len());
         self.nodes.push(NodeData {
             parent: Some(parent_id),
@@ -52,7 +49,8 @@ impl Document {
 
     fn append_attribute(&mut self, tag_name: EId, aid: AId, value: &str) {
         let value2 = parse_svg_attribute(tag_name, aid, value);
-        if let Ok(value) = value2 {
+        // TODO: improve error logging
+        if let Some(value) = value2 {
             self.attrs.push(Attribute {
                 name: aid,
                 value,
@@ -61,7 +59,7 @@ impl Document {
             // Invalid `enable-background` is not an error
             // since we are ignoring the `accumulate` value.
             if aid != AId::EnableBackground {
-                warn!("Failed to parse {} value: '{}'.", aid, value);
+                log::warn!("Failed to parse {} value: '{}'.", aid, value);
             }
         }
     }
@@ -114,7 +112,7 @@ fn parse(xml: &roxmltree::Document) -> Result<Document, Error> {
     Ok(doc)
 }
 
-fn parse_tag_name(node: roxmltree::Node) -> Option<EId> {
+pub(super) fn parse_tag_name(node: roxmltree::Node) -> Option<EId> {
     if !node.is_element() {
         return None;
     }
@@ -171,7 +169,10 @@ fn parse_xml_node(
 
     let node_id = parse_svg_element(node, parent_id, tag_name, style_sheet, ignore_ids, doc)?;
     if tag_name == EId::Text {
-        parse_svg_text_element(node, node_id, style_sheet, doc)?;
+        #[cfg(feature = "text")]
+        {
+            super::text::parse_svg_text_element(node, node_id, style_sheet, doc)?;
+        }
     } else if tag_name == EId::Use {
         parse_svg_use_element(node, origin, node_id, style_sheet, depth + 1, doc)?;
     } else {
@@ -181,7 +182,7 @@ fn parse_xml_node(
     Ok(())
 }
 
-fn parse_svg_element(
+pub(super) fn parse_svg_element(
     xml_node: roxmltree::Node,
     parent_id: NodeId,
     tag_name: EId,
@@ -201,7 +202,10 @@ fn parse_svg_element(
             _ => continue,
         }
 
-        let aid = try_opt_continue!(AId::from_str(attr.name()));
+        let aid = match AId::from_str(attr.name()) {
+            Some(v) => v,
+            None => continue,
+        };
 
         // During a `use` resolving, all `id` attributes must be ignored.
         // Otherwise we will get elements with duplicated id's.
@@ -310,12 +314,12 @@ fn parse_svg_attribute(
     tag_name: EId,
     aid: AId,
     value: &str,
-) -> Result<AttributeValue, svgtypes::Error> {
-    Ok(match aid {
+) -> Option<AttributeValue> {
+    Some(match aid {
         AId::Href => {
             // `href` can contain base64 data and we do store it as is.
-            match svgtypes::Stream::from(value).parse_iri() {
-                Ok(link) => AttributeValue::Link(link.to_string()),
+            match svgtypes::IRI::from_str(value) {
+                Ok(link) => AttributeValue::Link(link.0.to_string()),
                 Err(_) => AttributeValue::String(value.to_string()),
             }
         }
@@ -331,10 +335,10 @@ fn parse_svg_attribute(
                 }
                   EId::FePointLight
                 | EId::FeSpotLight => {
-                    AttributeValue::Number(parse_number(value)?)
+                    AttributeValue::Number(svgtypes::Number::from_str(value).ok()?.0)
                 }
                 _ => {
-                    AttributeValue::Length(svgtypes::Length::from_str(value)?)
+                    AttributeValue::Length(svgtypes::Length::from_str(value).ok()?)
                 }
             }
         }
@@ -349,26 +353,26 @@ fn parse_svg_attribute(
         | AId::Width | AId::Height
         | AId::MarkerWidth | AId::MarkerHeight
         | AId::StartOffset => {
-            AttributeValue::Length(svgtypes::Length::from_str(value)?)
+            AttributeValue::Length(svgtypes::Length::from_str(value).ok()?)
         }
 
         AId::Offset => {
             if let EId::FeFuncR | EId::FeFuncG | EId::FeFuncB | EId::FeFuncA = tag_name {
-                AttributeValue::Number(parse_number(value)?)
+                AttributeValue::Number(svgtypes::Number::from_str(value).ok()?.0)
             } else {
                 // offset = <number> | <percentage>
-                let l = svgtypes::Length::from_str(value)?;
+                let l = svgtypes::Length::from_str(value).ok()?;
                 if l.unit == svgtypes::LengthUnit::None || l.unit == svgtypes::LengthUnit::Percent {
                     AttributeValue::Length(l)
                 } else {
-                    return Err(svgtypes::Error::InvalidValue);
+                    return None;
                 }
             }
         }
 
           AId::StrokeDashoffset
         | AId::StrokeWidth => {
-            AttributeValue::Length(svgtypes::Length::from_str(value)?)
+            AttributeValue::Length(svgtypes::Length::from_str(value).ok()?)
         }
 
           AId::Opacity
@@ -376,7 +380,7 @@ fn parse_svg_attribute(
         | AId::FloodOpacity
         | AId::StrokeOpacity
         | AId::StopOpacity => {
-            let n = parse_number(value)?;
+            let n = svgtypes::Number::from_str(value).ok()?.0;
             let n = crate::utils::f64_bound(0.0, n, 1.0);
             AttributeValue::Opacity(n.into())
         }
@@ -408,7 +412,7 @@ fn parse_svg_attribute(
         | AId::TargetX
         | AId::TargetY
         | AId::Z => {
-            AttributeValue::Number(parse_number(value)?)
+            AttributeValue::Number(svgtypes::Number::from_str(value).ok()?.0)
         }
 
         AId::StrokeDasharray => {
@@ -428,14 +432,14 @@ fn parse_svg_attribute(
                     AttributeValue::Paint(link.to_string(), fallback)
                 }
                 Err(_) => {
-                    warn!("Failed to parse fill value: '{}'. Fallback to black.", value);
+                    log::warn!("Failed to parse fill value: '{}'. Fallback to black.", value);
                     AttributeValue::Color(svgtypes::Color::black())
                 }
             }
         }
 
         AId::Stroke => {
-            match svgtypes::Paint::from_str(value)? {
+            match svgtypes::Paint::from_str(value).ok()? {
                 svgtypes::Paint::None => AttributeValue::None,
                 svgtypes::Paint::Inherit => unreachable!(),
                 svgtypes::Paint::CurrentColor => AttributeValue::CurrentColor,
@@ -447,7 +451,6 @@ fn parse_svg_attribute(
         }
 
           AId::ClipPath
-        | AId::Filter
         | AId::MarkerEnd
         | AId::MarkerMid
         | AId::MarkerStart
@@ -455,15 +458,14 @@ fn parse_svg_attribute(
             match value {
                 "none" => AttributeValue::None,
                 _ => {
-                    let mut s = svgtypes::Stream::from(value);
-                    let link = s.parse_func_iri()?;
-                    AttributeValue::Link(link.to_string())
+                    let link = svgtypes::FuncIRI::from_str(value).ok()?;
+                    AttributeValue::Link(link.0.to_string())
                 }
             }
         }
 
         AId::Color => {
-            AttributeValue::Color(svgtypes::Color::from_str(value)?)
+            AttributeValue::Color(svgtypes::Color::from_str(value).ok()?)
         }
 
           AId::FloodColor
@@ -471,7 +473,7 @@ fn parse_svg_attribute(
         | AId::StopColor => {
             match value {
                 "currentColor" => AttributeValue::CurrentColor,
-                _ => AttributeValue::Color(svgtypes::Color::from_str(value)?),
+                _ => AttributeValue::Color(svgtypes::Color::from_str(value).ok()?),
             }
         }
 
@@ -480,14 +482,14 @@ fn parse_svg_attribute(
             if segments.len() >= 2 {
                 AttributeValue::Path(Rc::new(segments))
             } else {
-                return Err(svgtypes::Error::InvalidValue);
+                return None;
             }
         }
 
           AId::Transform
         | AId::GradientTransform
         | AId::PatternTransform => {
-            AttributeValue::Transform(svgtypes::Transform::from_str(value)?)
+            AttributeValue::Transform(svgtypes::Transform::from_str(value).ok()?.into())
         }
 
         AId::FontSize => {
@@ -508,30 +510,30 @@ fn parse_svg_attribute(
         | AId::WordSpacing => {
             match value {
                 "normal" => AttributeValue::String(value.to_string()),
-                _ => AttributeValue::Length(svgtypes::Length::from_str(value)?),
+                _ => AttributeValue::Length(svgtypes::Length::from_str(value).ok()?),
             }
         }
 
         AId::BaselineShift => {
             match value {
                 "baseline" | "sub" | "super" => AttributeValue::String(value.to_string()),
-                _ => AttributeValue::Length(svgtypes::Length::from_str(value)?),
+                _ => AttributeValue::Length(svgtypes::Length::from_str(value).ok()?),
             }
         }
 
         AId::Orient => {
             match value {
                 "auto" => AttributeValue::String(value.to_string()),
-                _ => AttributeValue::Angle(svgtypes::Angle::from_str(value)?),
+                _ => AttributeValue::Angle(svgtypes::Angle::from_str(value).ok()?),
             }
         }
 
         AId::ViewBox => {
-            AttributeValue::ViewBox(svgtypes::ViewBox::from_str(value)?)
+            AttributeValue::ViewBox(svgtypes::ViewBox::from_str(value).ok()?)
         }
 
         AId::PreserveAspectRatio => {
-            AttributeValue::AspectRatio(svgtypes::AspectRatio::from_str(value)?)
+            AttributeValue::AspectRatio(svgtypes::AspectRatio::from_str(value).ok()?)
         }
 
           AId::BaseFrequency
@@ -540,11 +542,28 @@ fn parse_svg_attribute(
         | AId::Rotate
         | AId::TableValues
         | AId::Values => {
-            AttributeValue::NumberList(svgtypes::NumberList::from_str(value)?)
+            let mut numbers = Vec::new();
+            for n in svgtypes::NumberListParser::from(value) {
+                numbers.push(n.ok()?);
+            }
+
+            AttributeValue::NumberList(numbers)
         }
 
         AId::EnableBackground => {
-            AttributeValue::EnableBackground(parse_enable_background(value)?)
+            let eb = svgtypes::EnableBackground::from_str(value).ok()?;
+            match eb {
+                svgtypes::EnableBackground::Accumulate => {
+                    return None
+                }
+                svgtypes::EnableBackground::New => {
+                    AttributeValue::EnableBackground(EnableBackground(None))
+                }
+                svgtypes::EnableBackground::NewWithRegion { x, y, width, height } => {
+                    let r = Rect::new(x, y, width, height)?;
+                    AttributeValue::EnableBackground(EnableBackground(Some(r)))
+                }
+            }
         }
 
         _ => {
@@ -554,19 +573,7 @@ fn parse_svg_attribute(
 }
 
 #[inline(never)]
-fn parse_number(value: &str) -> Result<f64, svgtypes::Error> {
-    let mut s = svgtypes::Stream::from(value);
-    let n = s.parse_number()?;
-
-    if !s.at_end() {
-        return Err(svgtypes::Error::InvalidNumber(0));
-    }
-
-    Ok(n)
-}
-
-#[inline(never)]
-fn parse_path(text: &str) -> tree::PathData {
+fn parse_path(text: &str) -> crate::PathData {
     // Previous MoveTo coordinates.
     let mut prev_mx = 0.0;
     let mut prev_my = 0.0;
@@ -581,7 +588,7 @@ fn parse_path(text: &str) -> tree::PathData {
 
     let mut prev_seg = svgtypes::PathSegment::MoveTo { abs: true, x: 0.0, y: 0.0 };
 
-    let mut path = tree::PathData::with_capacity(32);
+    let mut path = crate::PathData::with_capacity(32);
 
     for segment in svgtypes::PathParser::from(text) {
         let segment = match segment {
@@ -594,7 +601,7 @@ fn parse_path(text: &str) -> tree::PathData {
                 if !abs {
                     // When we get 'm'(relative) segment, which is not first segment - then it's
                     // relative to a previous 'M'(absolute) segment, not to the first segment.
-                    if let Some(tree::PathSegment::ClosePath) = path.last() {
+                    if let Some(crate::PathSegment::ClosePath) = path.last() {
                         x += prev_mx;
                         y += prev_my;
                     } else {
@@ -730,7 +737,7 @@ fn parse_path(text: &str) -> tree::PathData {
                 prev_seg = segment;
             }
             svgtypes::PathSegment::ClosePath { .. } => {
-                if let Some(tree::PathSegment::ClosePath) = path.last() {
+                if let Some(crate::PathSegment::ClosePath) = path.last() {
                     // Do not add sequential ClosePath segments.
                     // Otherwise it will break marker rendering.
                 } else {
@@ -744,21 +751,21 @@ fn parse_path(text: &str) -> tree::PathData {
         // Remember last position.
         if let Some(seg) = path.last() {
             match *seg {
-                tree::PathSegment::MoveTo { x, y } => {
+                crate::PathSegment::MoveTo { x, y } => {
                     prev_x = x;
                     prev_y = y;
                     prev_mx = x;
                     prev_my = y;
                 }
-                tree::PathSegment::LineTo { x, y } => {
+                crate::PathSegment::LineTo { x, y } => {
                     prev_x = x;
                     prev_y = y;
                 }
-                tree::PathSegment::CurveTo { x, y, .. } => {
+                crate::PathSegment::CurveTo { x, y, .. } => {
                     prev_x = x;
                     prev_y = y;
                 }
-                tree::PathSegment::ClosePath => {
+                crate::PathSegment::ClosePath => {
                     // ClosePath moves us to the last MoveTo coordinate,
                     // not previous.
                     prev_x = prev_mx;
@@ -772,33 +779,6 @@ fn parse_path(text: &str) -> tree::PathData {
     path
 }
 
-fn parse_enable_background(value: &str) -> Result<tree::EnableBackground, svgtypes::Error> {
-    let mut s = svgtypes::Stream::from(value);
-    s.skip_spaces();
-    if s.starts_with(b"new") {
-        s.advance(3);
-        s.skip_spaces();
-        if s.at_end() {
-            return Ok(tree::EnableBackground(None));
-        }
-
-        let x = s.parse_list_number()?;
-        let y = s.parse_list_number()?;
-        let w = s.parse_list_number()?;
-        let h = s.parse_list_number()?;
-
-        s.skip_spaces();
-        if !s.at_end() {
-            return Err(svgtypes::Error::InvalidValue);
-        }
-
-        let r = Rect::new(x, y, w, h).ok_or(svgtypes::Error::InvalidValue)?;
-        Ok(tree::EnableBackground(Some(r)))
-    } else {
-        Err(svgtypes::Error::InvalidValue)
-    }
-}
-
 fn resolve_inherit(
     parent_id: NodeId,
     tag_name: EId,
@@ -893,7 +873,7 @@ fn resolve_href<'a>(
     let link_value = node.attribute((XLINK_NS, "href"))
         .or_else(|| node.attribute("href"))?;
 
-    let link_id = svgtypes::Stream::from(link_value).parse_iri().ok()?;
+    let link_id = svgtypes::IRI::from_str(link_value).ok()?.0;
 
     // We're using `descendants` each time instead of HashTable because
     // we have to preserve the original elements order.
@@ -919,7 +899,7 @@ fn parse_svg_use_element(
     };
 
     if link == node || link == origin {
-        warn!("Recursive 'use' detected. '{}' will be skipped.",
+        log::warn!("Recursive 'use' detected. '{}' will be skipped.",
               node.attribute((SVG_NS, "id")).unwrap_or_default());
         return Ok(());
     }
@@ -932,7 +912,7 @@ fn parse_svg_use_element(
     // TODO: this
     // We don't support 'use' elements linked to 'svg' element.
     if tag_name == EId::Svg {
-        warn!("'use' elements linked to 'svg' elements are not supported. Skipped.");
+        log::warn!("'use' elements linked to 'svg' elements are not supported. Skipped.");
         return Ok(());
     }
 
@@ -964,355 +944,14 @@ fn parse_svg_use_element(
     }
 
     if is_recursive {
-        warn!("Recursive 'use' detected. '{}' will be skipped.",
-              node.attribute((SVG_NS, "id")).unwrap_or_default());
+        log::warn!("Recursive 'use' detected. '{}' will be skipped.",
+                   node.attribute((SVG_NS, "id")).unwrap_or_default());
         return Ok(());
     }
 
     parse_xml_node(link, node, parent_id, style_sheet, true, depth + 1, doc)
 }
 
-fn parse_svg_text_element(
-    parent: roxmltree::Node,
-    parent_id: NodeId,
-    style_sheet: &simplecss::StyleSheet,
-    doc: &mut Document,
-) -> Result<(), Error> {
-    debug_assert_eq!(parent.tag_name().name(), "text");
-
-    let space = if doc.get(parent_id).has_attribute(AId::Space) {
-        get_xmlspace(doc, parent_id, XmlSpace::Default)
-    } else {
-        if let Some(node) = doc.get(parent_id).ancestors().find(|n| n.has_attribute(AId::Space)) {
-            get_xmlspace(doc, node.id, XmlSpace::Default)
-        } else {
-            XmlSpace::Default
-        }
-    };
-
-    parse_svg_text_element_impl(parent, parent_id, style_sheet, space, doc)?;
-
-    trim_text_nodes(parent_id, space, doc);
-    Ok(())
-}
-
-fn parse_svg_text_element_impl(
-    parent: roxmltree::Node,
-    parent_id: NodeId,
-    style_sheet: &simplecss::StyleSheet,
-    space: XmlSpace,
-    doc: &mut Document,
-) -> Result<(), Error> {
-    for node in parent.children() {
-        if node.is_text() {
-            let text = trim_text(node.text().unwrap(), space);
-            doc.append(parent_id, NodeKind::Text(text));
-            continue;
-        }
-
-        let mut tag_name = try_opt_continue!(parse_tag_name(node));
-
-        if tag_name == EId::A {
-            // Treat links as simple text.
-            tag_name = EId::Tspan;
-        }
-
-        if !matches!(tag_name, EId::Tspan | EId::Tref | EId::TextPath) {
-            continue;
-        }
-
-        // `textPath` must be a direct `text` child.
-        if tag_name == EId::TextPath && parent.tag_name().name() != "text" {
-            continue;
-        }
-
-        // We are converting `tref` into `tspan` to simplify later use.
-        let mut is_tref = false;
-        if tag_name == EId::Tref {
-            tag_name = EId::Tspan;
-            is_tref = true;
-        }
-
-        let node_id = parse_svg_element(node, parent_id, tag_name, style_sheet, false, doc)?;
-        let space = get_xmlspace(doc, node_id, space);
-
-        if is_tref {
-            let link_value = node.attribute((XLINK_NS, "href"))
-                .or_else(|| node.attribute("href"));
-
-            if let Some(href) = link_value {
-                if let Some(text) = resolve_tref_text(node.document(), href) {
-                    let text = trim_text(&text, space);
-                    doc.append(node_id, NodeKind::Text(text));
-                }
-            }
-        } else {
-            parse_svg_text_element_impl(node, node_id, style_sheet, space, doc)?;
-        }
-    }
-
-    Ok(())
-}
-
-fn resolve_tref_text(
-    xml: &roxmltree::Document,
-    href: &str,
-) -> Option<String> {
-    let id = svgtypes::Stream::from(href).parse_iri().ok()?;
-
-    // Find linked element in the original tree.
-    let node = xml.descendants().find(|n| n.attribute("id") == Some(id))?;
-
-    // `tref` should be linked to an SVG element.
-    parse_tag_name(node)?;
-
-    // 'All character data within the referenced element, including character data enclosed
-    // within additional markup, will be rendered.'
-    //
-    // So we don't care about attributes and everything. Just collecting text nodes data.
-    //
-    // Note: we have to filter nodes by `is_text()` first since `text()` will look up
-    // for text nodes in element children therefore we will get duplicates.
-    let text: String = node.descendants().filter(|n| n.is_text()).filter_map(|n| n.text()).collect();
-    if text.is_empty() { None } else { Some(text) }
-}
-
-#[derive(Clone, Copy, PartialEq, Debug)]
-enum XmlSpace {
-    Default,
-    Preserve,
-}
-
-fn get_xmlspace(doc: &Document, node_id: NodeId, default: XmlSpace) -> XmlSpace {
-    match doc.get(node_id).attribute(AId::Space) {
-        Some("preserve") => XmlSpace::Preserve,
-        Some(_) => XmlSpace::Default,
-        _ => default,
-    }
-}
-
-trait StrTrim {
-    fn remove_first_space(&mut self);
-    fn remove_last_space(&mut self);
-}
-
-impl StrTrim for String {
-    fn remove_first_space(&mut self) {
-        debug_assert_eq!(self.chars().next().unwrap(), ' ');
-        self.drain(0..1);
-    }
-
-    fn remove_last_space(&mut self) {
-        debug_assert_eq!(self.chars().rev().next().unwrap(), ' ');
-        self.pop();
-    }
-}
-
-/// Prepares text nodes according to the spec: https://www.w3.org/TR/SVG11/text.html#WhiteSpace
-///
-/// This function handles:
-/// - 'xml:space' processing
-/// - tabs and newlines removing/replacing
-/// - spaces trimming
-fn trim_text_nodes(text_elem_id: NodeId, xmlspace: XmlSpace, doc: &mut Document) {
-    let mut nodes = Vec::new(); // TODO: allocate only once
-    collect_text_nodes(doc.get(text_elem_id), 0, &mut nodes);
-
-    // `trim` method has already collapsed all spaces into a single one,
-    // so we have to check only for one leading or trailing space.
-
-    if nodes.len() == 1 {
-        // Process element with a single text node child.
-
-        let node_id = nodes[0].0;
-
-        if xmlspace == XmlSpace::Default {
-            if let NodeKind::Text(ref mut text) = doc.nodes[node_id.0].kind {
-                match text.len() {
-                    0 => {} // An empty string. Do nothing.
-                    1 => {
-                        // If string has only one character and it's a space - clear this string.
-                        if text.as_bytes()[0] == b' ' {
-                            text.clear();
-                        }
-                    }
-                    _ => {
-                        // 'text' has at least 2 bytes, so indexing is safe.
-                        let c1 = text.as_bytes()[0];
-                        let c2 = text.as_bytes()[text.len() - 1];
-
-                        if c1 == b' ' {
-                            text.remove_first_space();
-                        }
-
-                        if c2 == b' ' {
-                            text.remove_last_space();
-                        }
-                    }
-                }
-            }
-        } else {
-            // Do nothing when xml:space=preserve.
-        }
-    } else if nodes.len() > 1 {
-        // Process element with many text node children.
-
-        // We manage all text nodes as a single text node
-        // and trying to remove duplicated spaces across nodes.
-        //
-        // For example    '<text>Text <tspan> text </tspan> text</text>'
-        // is the same is '<text>Text <tspan>text</tspan> text</text>'
-
-        let mut i = 0;
-        let len = nodes.len() - 1;
-        let mut last_non_empty: Option<NodeId> = None;
-        while i < len {
-            // Process pairs.
-            let (mut node1_id, depth1) = nodes[i];
-            let (node2_id, depth2) = nodes[i + 1];
-
-            if doc.get(node1_id).text().is_empty() {
-                if let Some(n) = last_non_empty {
-                    node1_id = n;
-                }
-            }
-
-            // Parent of the text node is always an element node and always exist,
-            // so unwrap is safe.
-            let xmlspace1 = get_xmlspace(doc, doc.get(node1_id).parent().unwrap().id, xmlspace);
-            let xmlspace2 = get_xmlspace(doc, doc.get(node2_id).parent().unwrap().id, xmlspace);
-
-            // >text<..>text<
-            //  1  2    3  4
-            let (c1, c2, c3, c4) = {
-                let text1 = doc.get(node1_id).text();
-                let text2 = doc.get(node2_id).text();
-
-                let bytes1 = text1.as_bytes();
-                let bytes2 = text2.as_bytes();
-
-                let c1 = bytes1.first().cloned();
-                let c2 = bytes1.last().cloned();
-                let c3 = bytes2.first().cloned();
-                let c4 = bytes2.last().cloned();
-
-                (c1, c2, c3, c4)
-            };
-
-            // NOTE: xml:space processing is mostly an undefined behavior,
-            // because everyone do it differently.
-            // We're mimicking the Chrome behavior.
-
-            // Remove space from the second text node if both nodes has bound spaces.
-            // From: '<text>Text <tspan> text</tspan></text>'
-            // To:   '<text>Text <tspan>text</tspan></text>'
-            //
-            // See text-tspan-02-b.svg for details.
-            if depth1 < depth2 {
-                if c3 == Some(b' ') {
-                    if xmlspace2 == XmlSpace::Default {
-                        if let NodeKind::Text(ref mut text) = doc.nodes[node2_id.0].kind {
-                            text.remove_first_space();
-                        }
-                    }
-                }
-            } else {
-                if c2 == Some(b' ') && c2 == c3 {
-                    if xmlspace1 == XmlSpace::Default && xmlspace2 == XmlSpace::Default {
-                        if let NodeKind::Text(ref mut text) = doc.nodes[node1_id.0].kind {
-                            text.remove_last_space();
-                        }
-                    } else {
-                        if xmlspace1 == XmlSpace::Preserve && xmlspace2 == XmlSpace::Default {
-                            if let NodeKind::Text(ref mut text) = doc.nodes[node2_id.0].kind {
-                                text.remove_first_space();
-                            }
-                        }
-                    }
-                }
-            }
-
-            let is_first = i == 0;
-            let is_last  = i == len - 1;
-
-            if     is_first
-                && c1 == Some(b' ')
-                && xmlspace1 == XmlSpace::Default
-                && !doc.get(node1_id).text().is_empty()
-            {
-                // Remove a leading space from a first text node.
-                if let NodeKind::Text(ref mut text) = doc.nodes[node1_id.0].kind {
-                    text.remove_first_space();
-                }
-            } else if
-                   is_last
-                && c4 == Some(b' ')
-                && !doc.get(node2_id).text().is_empty()
-                && xmlspace2 == XmlSpace::Default
-            {
-                // Remove a trailing space from a last text node.
-                // Also check that 'text2' is not empty already.
-                if let NodeKind::Text(ref mut text) = doc.nodes[node2_id.0].kind {
-                    text.remove_last_space();
-                }
-            }
-
-            if     is_last
-                && c2 == Some(b' ')
-                && !doc.get(node1_id).text().is_empty()
-                && doc.get(node2_id).text().is_empty()
-                && doc.get(node1_id).text().ends_with(' ')
-            {
-                if let NodeKind::Text(ref mut text) = doc.nodes[node1_id.0].kind {
-                    text.remove_last_space();
-                }
-            }
-
-            if !doc.get(node1_id).text().trim().is_empty() {
-                last_non_empty = Some(node1_id);
-            }
-
-            i += 1;
-        }
-    }
-
-    // TODO: find a way to remove all empty text nodes
-}
-
-fn collect_text_nodes(parent: Node, depth: usize, nodes: &mut Vec<(NodeId, usize)>) {
-    for child in parent.children() {
-        if child.is_text() {
-            nodes.push((child.id, depth));
-        } else if child.is_element() {
-            collect_text_nodes(child, depth + 1, nodes);
-        }
-    }
-}
-
-fn trim_text(text: &str, space: XmlSpace) -> String {
-    let mut s = String::with_capacity(text.len());
-
-    let mut prev = '0';
-    for c in text.chars() {
-        // \r, \n and \t should be converted into spaces.
-        let c = match c {
-            '\r' | '\n' | '\t' => ' ',
-            _ => c,
-        };
-
-        // Skip continuous spaces.
-        if space == XmlSpace::Default && c == ' ' && c == prev {
-            continue;
-        }
-
-        prev = c;
-
-        s.push(c);
-    }
-
-    s
-}
-
 fn resolve_css<'a>(xml: &'a roxmltree::Document<'a>) -> simplecss::StyleSheet<'a> {
     let mut sheet = simplecss::StyleSheet::new();
 
@@ -1323,7 +962,12 @@ fn resolve_css<'a>(xml: &'a roxmltree::Document<'a>) -> simplecss::StyleSheet<'a
             None => {}
         }
 
-        sheet.parse_more(try_opt_continue!(node.text()));
+        let text = match node.text() {
+            Some(v) => v,
+            None => continue,
+        };
+
+        sheet.parse_more(text);
     }
 
     sheet
